@@ -1,10 +1,8 @@
-/* $Id$ */
-
 /*
  * This file is part of OpenTTD.
  * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
  * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
- * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
+ * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <https://www.gnu.org/licenses/old-licenses/gpl-2.0>.
  */
 
 /** @file win32_m.cpp Music playback for Windows. */
@@ -19,26 +17,28 @@
 #include "midifile.hpp"
 #include "midi.h"
 #include "../base_media_base.h"
+#include "../base_media_music.h"
+#include <mutex>
 
 #include "../safeguards.h"
 
 struct PlaybackSegment {
-	uint32 start, end;
+	uint32_t start, end;
 	size_t start_block;
 	bool loop;
 };
 
 static struct {
-	UINT time_period;      ///< obtained timer precision value
-	HMIDIOUT midi_out;     ///< handle to open midiOut
-	UINT timer_id;         ///< ID of active multimedia timer
-	CRITICAL_SECTION lock; ///< synchronization for playback status fields
+	UINT time_period;    ///< obtained timer precision value
+	HMIDIOUT midi_out;   ///< handle to open midiOut
+	UINT timer_id;       ///< ID of active multimedia timer
+	std::mutex lock;     ///< synchronization for playback status fields
 
 	bool playing;        ///< flag indicating that playback is active
-	bool do_start;       ///< flag for starting playback of next_file at next opportunity
+	int do_start;        ///< flag for starting playback of next_file at next opportunity
 	bool do_stop;        ///< flag for stopping playback at next opportunity
-	byte current_volume; ///< current effective volume setting
-	byte new_volume;     ///< volume setting to change to
+	uint8_t current_volume; ///< current effective volume setting
+	uint8_t new_volume;     ///< volume setting to change to
 
 	MidiFile current_file;           ///< file currently being played from
 	PlaybackSegment current_segment; ///< segment info for current playback
@@ -47,49 +47,47 @@ static struct {
 	MidiFile next_file;              ///< upcoming file to play
 	PlaybackSegment next_segment;    ///< segment info for upcoming file
 
-	byte channel_volumes[16]; ///< last seen volume controller values in raw data
+	std::array<uint8_t, 16> channel_volumes; ///< last seen volume controller values in raw data
 } _midi;
 
 static FMusicDriver_Win32 iFMusicDriver_Win32;
 
 
-static byte ScaleVolume(byte original, byte scale)
+static uint8_t ScaleVolume(uint8_t original, uint8_t scale)
 {
 	return original * scale / 127;
 }
 
 
-void CALLBACK MidiOutProc(HMIDIOUT hmo, UINT wMsg, DWORD_PTR dwInstance, DWORD_PTR dwParam1, DWORD_PTR dwParam2)
+void CALLBACK MidiOutProc(HMIDIOUT hmo, UINT wMsg, DWORD_PTR, DWORD_PTR dwParam1, DWORD_PTR)
 {
 	if (wMsg == MOM_DONE) {
 		MIDIHDR *hdr = (LPMIDIHDR)dwParam1;
 		midiOutUnprepareHeader(hmo, hdr, sizeof(*hdr));
-		free(hdr);
+		delete hdr;
 	}
 }
 
-static void TransmitChannelMsg(byte status, byte p1, byte p2 = 0)
+static void TransmitChannelMsg(uint8_t status, uint8_t p1, uint8_t p2 = 0)
 {
 	midiOutShortMsg(_midi.midi_out, status | (p1 << 8) | (p2 << 16));
 }
 
-static void TransmitSysex(byte *&msg_start, size_t &remaining)
+static void TransmitSysex(const uint8_t *&msg_start, size_t &remaining)
 {
 	/* find end of message */
-	byte *msg_end = msg_start;
+	const uint8_t *msg_end = msg_start;
 	while (*msg_end != MIDIST_ENDSYSEX) msg_end++;
 	msg_end++; /* also include sysex end byte */
 
 	/* prepare header */
-	MIDIHDR *hdr = CallocT<MIDIHDR>(1);
-	hdr->lpData = (LPSTR)msg_start;
-	hdr->dwBufferLength = msg_end - msg_start;
-	if (midiOutPrepareHeader(_midi.midi_out, hdr, sizeof(*hdr)) == MMSYSERR_NOERROR) {
+	auto hdr = std::make_unique<MIDIHDR>();
+	hdr->lpData = reinterpret_cast<LPSTR>(const_cast<uint8_t *>(msg_start));
+	hdr->dwBufferLength = static_cast<DWORD>(msg_end - msg_start);
+	if (midiOutPrepareHeader(_midi.midi_out, hdr.get(), sizeof(MIDIHDR)) == MMSYSERR_NOERROR) {
 		/* transmit - just point directly into the data buffer */
 		hdr->dwBytesRecorded = hdr->dwBufferLength;
-		midiOutLongMsg(_midi.midi_out, hdr, sizeof(*hdr));
-	} else {
-		free(hdr);
+		midiOutLongMsg(_midi.midi_out, hdr.release(), sizeof(MIDIHDR));
 	}
 
 	/* update position in buffer */
@@ -97,100 +95,113 @@ static void TransmitSysex(byte *&msg_start, size_t &remaining)
 	msg_start = msg_end;
 }
 
-static void TransmitSysexConst(byte *msg_start, size_t length)
+static void TransmitStandardSysex(MidiSysexMessage msg)
 {
-	TransmitSysex(msg_start, length);
+	size_t length = 0;
+	const uint8_t *data = MidiGetStandardSysexMessage(msg, length);
+	TransmitSysex(data, length);
 }
 
 /**
  * Realtime MIDI playback service routine.
  * This is called by the multimedia timer.
  */
-void CALLBACK TimerCallback(UINT uTimerID, UINT, DWORD_PTR dwUser, DWORD_PTR, DWORD_PTR)
+void CALLBACK TimerCallback(UINT uTimerID, UINT, DWORD_PTR, DWORD_PTR, DWORD_PTR)
 {
-	/* Try to check playback status changes.
-	 * If _midi is already locked, skip checking for this cycle and try again
-	 * next cycle, instead of waiting for locks in the realtime callback. */
-	if (TryEnterCriticalSection(&_midi.lock)) {
-		/* check for stop */
-		if (_midi.do_stop) {
-			DEBUG(driver, 2, "Win32-MIDI: timer: do_stop is set");
-			midiOutReset(_midi.midi_out);
-			_midi.playing = false;
-			_midi.do_stop = false;
-			LeaveCriticalSection(&_midi.lock);
+	static int volume_throttle = 0;
+
+	/* Ensure only one timer callback is running at once, and prevent races on status flags */
+	std::unique_lock<std::mutex> mutex_lock(_midi.lock, std::defer_lock);
+	if (!mutex_lock.try_lock()) return;
+
+	/* check for stop */
+	if (_midi.do_stop) {
+		Debug(driver, 2, "Win32-MIDI: timer: do_stop is set");
+		midiOutReset(_midi.midi_out);
+		_midi.playing = false;
+		_midi.do_stop = false;
+		return;
+	}
+
+	/* check for start/restart/change song */
+	if (_midi.do_start != 0) {
+		/* Have a delay between playback start steps, prevents jumbled-together notes at the start of song */
+		if (timeGetTime() - _midi.playback_start_time < 50) {
 			return;
 		}
+		Debug(driver, 2, "Win32-MIDI: timer: do_start step {}", _midi.do_start);
 
-		/* check for start/restart/change song */
-		if (_midi.do_start) {
-			DEBUG(driver, 2, "Win32-MIDI: timer: do_start is set");
-			if (_midi.playing) {
-				midiOutReset(_midi.midi_out);
-				/* Some songs change the "Pitch bend range" registered
-				 * parameter. If this doesn't get reset, everything else
-				 * will start sounding wrong. */
-				for (int ch = 0; ch < 16; ch++) {
-					/* Running status, only need status for first message */
-					/* Select RPN 00.00, set value to 02.00, and unselect again */
-					TransmitChannelMsg(MIDIST_CONTROLLER | ch, MIDICT_RPN_SELECT_LO, 0x00);
-					TransmitChannelMsg(MIDICT_RPN_SELECT_HI, 0x00);
-					TransmitChannelMsg(MIDICT_DATAENTRY,     0x02);
-					TransmitChannelMsg(MIDICT_DATAENTRY_LO,  0x00);
-					TransmitChannelMsg(MIDICT_RPN_SELECT_LO, 0x7F);
-					TransmitChannelMsg(MIDICT_RPN_SELECT_HI, 0x7F);
-				}
-			}
+		if (_midi.do_start == 1) {
+			/* Send "all notes off" */
+			midiOutReset(_midi.midi_out);
+			_midi.playback_start_time = timeGetTime();
+			_midi.do_start = 2;
+
+			return;
+		} else if (_midi.do_start == 2) {
+			/* Reset the device to General MIDI defaults */
+			TransmitStandardSysex(MidiSysexMessage::ResetGM);
+			_midi.playback_start_time = timeGetTime();
+			_midi.do_start = 3;
+
+			return;
+		} else if (_midi.do_start == 3) {
+			/* Set up device-specific effects */
+			TransmitStandardSysex(MidiSysexMessage::RolandSetReverb);
+			_midi.playback_start_time = timeGetTime();
+			_midi.do_start = 4;
+
+			return;
+		} else if (_midi.do_start == 4) {
+			/* Load the new file */
 			_midi.current_file.MoveFrom(_midi.next_file);
 			std::swap(_midi.next_segment, _midi.current_segment);
 			_midi.current_segment.start_block = 0;
 			_midi.playback_start_time = timeGetTime();
 			_midi.playing = true;
-			_midi.do_start = false;
+			_midi.do_start = 0;
 			_midi.current_block = 0;
 
-			MemSetT<byte>(_midi.channel_volumes, 127, lengthof(_midi.channel_volumes));
-		} else if (!_midi.playing) {
-			/* not playing, stop the timer */
-			DEBUG(driver, 2, "Win32-MIDI: timer: not playing, stopping timer");
-			timeKillEvent(uTimerID);
-			_midi.timer_id = 0;
-			LeaveCriticalSection(&_midi.lock);
-			return;
+			_midi.channel_volumes.fill(127);
+			/* Invalidate current volume. */
+			_midi.current_volume = UINT8_MAX;
+			volume_throttle = 0;
 		}
+	} else if (!_midi.playing) {
+		/* not playing, stop the timer */
+		Debug(driver, 2, "Win32-MIDI: timer: not playing, stopping timer");
+		timeKillEvent(uTimerID);
+		_midi.timer_id = 0;
+		return;
+	}
 
-		/* check for volume change */
-		static int volume_throttle = 0;
-		if (_midi.current_volume != _midi.new_volume) {
-			if (volume_throttle == 0) {
-				DEBUG(driver, 2, "Win32-MIDI: timer: volume change");
-				_midi.current_volume = _midi.new_volume;
-				volume_throttle = 20 / _midi.time_period;
-				for (int ch = 0; ch < 16; ch++) {
-					int vol = ScaleVolume(_midi.channel_volumes[ch], _midi.current_volume);
-					TransmitChannelMsg(MIDIST_CONTROLLER | ch, MIDICT_CHANVOLUME, vol);
-				}
-			}
-			else {
-				volume_throttle--;
+	/* check for volume change */
+	if (_midi.current_volume != _midi.new_volume) {
+		if (volume_throttle == 0) {
+			Debug(driver, 2, "Win32-MIDI: timer: volume change");
+			_midi.current_volume = _midi.new_volume;
+			volume_throttle = 20 / _midi.time_period;
+			for (int ch = 0; ch < 16; ch++) {
+				uint8_t vol = ScaleVolume(_midi.channel_volumes[ch], _midi.current_volume);
+				TransmitChannelMsg(MIDIST_CONTROLLER | ch, MIDICT_CHANVOLUME, vol);
 			}
+		} else {
+			volume_throttle--;
 		}
-
-		LeaveCriticalSection(&_midi.lock);
 	}
 
 	/* skip beginning of file? */
 	if (_midi.current_segment.start > 0 && _midi.current_block == 0 && _midi.current_segment.start_block == 0) {
 		/* find first block after start time and pretend playback started earlier
-		* this is to allow all blocks prior to the actual start to still affect playback,
-		* as they may contain important controller and program changes */
-		uint preload_bytes = 0;
+		 * this is to allow all blocks prior to the actual start to still affect playback,
+		 * as they may contain important controller and program changes */
+		size_t preload_bytes = 0;
 		for (size_t bl = 0; bl < _midi.current_file.blocks.size(); bl++) {
 			MidiFile::DataBlock &block = _midi.current_file.blocks[bl];
-			preload_bytes += block.data.Length();
+			preload_bytes += block.data.size();
 			if (block.ticktime >= _midi.current_segment.start) {
 				if (_midi.current_segment.loop) {
-					DEBUG(driver, 2, "Win32-MIDI: timer: loop from block %d (ticktime %d, realtime %.3f, bytes %d)", (int)bl, (int)block.ticktime, ((int)block.realtime)/1000.0, (int)preload_bytes);
+					Debug(driver, 2, "Win32-MIDI: timer: loop from block {} (ticktime {}, realtime {:.3f}, bytes {})", bl, block.ticktime, block.realtime / 1000.0, preload_bytes);
 					_midi.current_segment.start_block = bl;
 					break;
 				} else {
@@ -199,8 +210,8 @@ void CALLBACK TimerCallback(UINT uTimerID, UINT, DWORD_PTR dwUser, DWORD_PTR, DW
 					 * which have a bitrate of 31,250 bits/sec, and transmit 1+8+1 start/data/stop bits per byte.
 					 * The delay compensation is needed to avoid time-compression of following messages.
 					 */
-					DEBUG(driver, 2, "Win32-MIDI: timer: start from block %d (ticktime %d, realtime %.3f, bytes %d)", (int)bl, (int)block.ticktime, ((int)block.realtime) / 1000.0, (int)preload_bytes);
-					_midi.playback_start_time -= block.realtime / 1000 - preload_bytes * 1000 / 3125;
+					Debug(driver, 2, "Win32-MIDI: timer: start from block {} (ticktime {}, realtime {:.3f}, bytes {})", bl, block.ticktime, block.realtime / 1000.0, preload_bytes);
+					_midi.playback_start_time -= block.realtime / 1000 - (DWORD)(preload_bytes * 1000 / 3125);
 					break;
 				}
 			}
@@ -229,13 +240,13 @@ void CALLBACK TimerCallback(UINT uTimerID, UINT, DWORD_PTR dwUser, DWORD_PTR, DW
 			break;
 		}
 
-		byte *data = block.data.Begin();
-		size_t remaining = block.data.Length();
-		byte last_status = 0;
+		const uint8_t *data = block.data.data();
+		size_t remaining = block.data.size();
+		uint8_t last_status = 0;
 		while (remaining > 0) {
 			/* MidiFile ought to have converted everything out of running status,
 			 * but handle it anyway just to be safe */
-			byte status = data[0];
+			uint8_t status = data[0];
 			if (status & 0x80) {
 				last_status = status;
 				data++;
@@ -312,62 +323,77 @@ void CALLBACK TimerCallback(UINT uTimerID, UINT, DWORD_PTR dwUser, DWORD_PTR, DW
 
 void MusicDriver_Win32::PlaySong(const MusicSongInfo &song)
 {
-	DEBUG(driver, 2, "Win32-MIDI: PlaySong: entry");
-	EnterCriticalSection(&_midi.lock);
+	Debug(driver, 2, "Win32-MIDI: PlaySong: entry");
 
-	if (!_midi.next_file.LoadSong(song)) {
-		LeaveCriticalSection(&_midi.lock);
-		return;
-	}
+	MidiFile new_song;
+	if (!new_song.LoadSong(song)) return;
+	Debug(driver, 2, "Win32-MIDI: PlaySong: Loaded song");
 
+	std::lock_guard<std::mutex> mutex_lock(_midi.lock);
+
+	_midi.next_file.MoveFrom(new_song);
 	_midi.next_segment.start = song.override_start;
 	_midi.next_segment.end = song.override_end;
 	_midi.next_segment.loop = song.loop;
 
-	DEBUG(driver, 2, "Win32-MIDI: PlaySong: setting flag");
+	Debug(driver, 2, "Win32-MIDI: PlaySong: setting flag");
 	_midi.do_stop = _midi.playing;
-	_midi.do_start = true;
+	_midi.do_start = 1;
 
 	if (_midi.timer_id == 0) {
-		DEBUG(driver, 2, "Win32-MIDI: PlaySong: starting timer");
+		Debug(driver, 2, "Win32-MIDI: PlaySong: starting timer");
 		_midi.timer_id = timeSetEvent(_midi.time_period, _midi.time_period, TimerCallback, (DWORD_PTR)this, TIME_PERIODIC | TIME_CALLBACK_FUNCTION);
 	}
-
-	LeaveCriticalSection(&_midi.lock);
 }
 
 void MusicDriver_Win32::StopSong()
 {
-	DEBUG(driver, 2, "Win32-MIDI: StopSong: entry");
-	EnterCriticalSection(&_midi.lock);
-	DEBUG(driver, 2, "Win32-MIDI: StopSong: setting flag");
+	Debug(driver, 2, "Win32-MIDI: StopSong: entry");
+	std::lock_guard<std::mutex> mutex_lock(_midi.lock);
+	Debug(driver, 2, "Win32-MIDI: StopSong: setting flag");
 	_midi.do_stop = true;
-	LeaveCriticalSection(&_midi.lock);
 }
 
 bool MusicDriver_Win32::IsSongPlaying()
 {
-	return _midi.playing || _midi.do_start;
+	return _midi.playing || (_midi.do_start != 0);
 }
 
-void MusicDriver_Win32::SetVolume(byte vol)
+void MusicDriver_Win32::SetVolume(uint8_t vol)
 {
-	EnterCriticalSection(&_midi.lock);
+	std::lock_guard<std::mutex> mutex_lock(_midi.lock);
 	_midi.new_volume = vol;
-	LeaveCriticalSection(&_midi.lock);
 }
 
-const char *MusicDriver_Win32::Start(const char * const *parm)
+std::optional<std::string_view> MusicDriver_Win32::Start(const StringList &parm)
 {
-	DEBUG(driver, 2, "Win32-MIDI: Start: initializing");
-
-	InitializeCriticalSection(&_midi.lock);
+	Debug(driver, 2, "Win32-MIDI: Start: initializing");
 
 	int resolution = GetDriverParamInt(parm, "resolution", 5);
-	int port = GetDriverParamInt(parm, "port", -1);
+	uint port = (uint)GetDriverParamInt(parm, "port", UINT_MAX);
+	auto portname = GetDriverParam(parm, "portname");
+
+	/* Enumerate ports either for selecting port by name, or for debug output */
+	if (portname.has_value() || _debug_driver_level > 0) {
+		uint numports = midiOutGetNumDevs();
+		Debug(driver, 1, "Win32-MIDI: Found {} output devices:", numports);
+		for (uint tryport = 0; tryport < numports; tryport++) {
+			MIDIOUTCAPS moc{};
+			if (midiOutGetDevCaps(tryport, &moc, sizeof(moc)) == MMSYSERR_NOERROR) {
+				char tryportname[128];
+				convert_from_fs(moc.szPname, tryportname);
+
+				/* Compare requested and detected port name.
+				 * If multiple ports have the same name, this will select the last matching port, and the debug output will be confusing. */
+				if (portname.has_value() && *portname == tryportname) port = tryport;
+
+				Debug(driver, 1, "MIDI port {:2d}: {}{}", tryport, tryportname, (tryport == port) ? " [selected]" : "");
+			}
+		}
+	}
 
 	UINT devid;
-	if (port < 0) {
+	if (port == UINT_MAX) {
 		devid = MIDI_MAPPER;
 	} else {
 		devid = (UINT)port;
@@ -381,22 +407,14 @@ const char *MusicDriver_Win32::Start(const char * const *parm)
 
 	midiOutReset(_midi.midi_out);
 
-	/* Standard "Enable General MIDI" message */
-	static byte gm_enable_sysex[] = { 0xF0, 0x7E, 0x00, 0x09, 0x01, 0xF7 };
-	TransmitSysexConst(&gm_enable_sysex[0], sizeof(gm_enable_sysex));
-
-	/* Roland-specific reverb room control, used by the original game */
-	static byte roland_reverb_sysex[] = { 0xF0, 0x41, 0x10, 0x42, 0x12, 0x40, 0x01, 0x30, 0x02, 0x04, 0x00, 0x40, 0x40, 0x00, 0x00, 0x09, 0xF7 };
-	TransmitSysexConst(&roland_reverb_sysex[0], sizeof(roland_reverb_sysex));
-
 	/* prepare multimedia timer */
 	TIMECAPS timecaps;
 	if (timeGetDevCaps(&timecaps, sizeof(timecaps)) == MMSYSERR_NOERROR) {
-		_midi.time_period = min(max((UINT)resolution, timecaps.wPeriodMin), timecaps.wPeriodMax);
+		_midi.time_period = std::min(std::max((UINT)resolution, timecaps.wPeriodMin), timecaps.wPeriodMax);
 		if (timeBeginPeriod(_midi.time_period) == MMSYSERR_NOERROR) {
 			/* success */
-			DEBUG(driver, 2, "Win32-MIDI: Start: timer resolution is %d", (int)_midi.time_period);
-			return NULL;
+			Debug(driver, 2, "Win32-MIDI: Start: timer resolution is {}", _midi.time_period);
+			return std::nullopt;
 		}
 	}
 	midiOutClose(_midi.midi_out);
@@ -405,7 +423,7 @@ const char *MusicDriver_Win32::Start(const char * const *parm)
 
 void MusicDriver_Win32::Stop()
 {
-	EnterCriticalSection(&_midi.lock);
+	std::lock_guard<std::mutex> mutex_lock(_midi.lock);
 
 	if (_midi.timer_id) {
 		timeKillEvent(_midi.timer_id);
@@ -415,7 +433,4 @@ void MusicDriver_Win32::Stop()
 	timeEndPeriod(_midi.time_period);
 	midiOutReset(_midi.midi_out);
 	midiOutClose(_midi.midi_out);
-
-	LeaveCriticalSection(&_midi.lock);
-	DeleteCriticalSection(&_midi.lock);
 }
